# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/
import requests
from pyincore import globals as pyglobals

logger = pyglobals.LOGGER

def return_http_response(http_response):
    try:
        http_response.raise_for_status()
        return http_response
    except requests.exceptions.HTTPError:
        logger.error('A HTTPError has occurred \n' +
                     'HTTP Status code: ' + str(http_response.status_code) + '\n' +
                     'Error Message: ' + http_response.content.decode()
                     )
        raise
    except requests.exceptions.ConnectionError:
        logger.error("ConnectionError: Failed to establish a connection with the server. "
                     "This might be due to a refused connection. "
                     "Please check that you are using the right URLs.")
        raise
    except requests.exceptions.RequestException:
        logger.error("RequestException: There was an exception while trying to handle your request. "
                     "Please go to the end of this message for more specific information about the exception.")
        raise



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/
import json
import os


from pyincore.dataservice import DataService

from pyincore import Dataset
import networkx as nx


class NetworkDataset:
    """ This class wraps around the Dataset class.

    Args:
        dataset (obj): The dataset object we want to extract the network data from.

    """

    def __init__(self, dataset: Dataset):
        if dataset.format == 'shp-network' and dataset.metadata['networkDataset'] is not None:
            self.metadata = dataset.metadata
            self.data_type = dataset.metadata["dataType"]
            self.nodes = NetworkDataset._network_component_from_dataset(dataset, "node")
            self.links = NetworkDataset._network_component_from_dataset(dataset, "link")
            self.graph = NetworkDataset._network_component_from_dataset(dataset, "graph")
        else:
            # TODO why do we need those
            self._metadata = None
            self._data_type = None
            self._links = None
            self._nodes = None
            self._graph = None

    @classmethod
    def from_data_service(cls, id: str, data_service: DataService):
        """Get Dataset from Data service, get metadata as well.

        Args:
            id (str): ID of the Dataset.
            data_service (obj): Data service.

        Returns:
            obj: network dataset

        """
        metadata = data_service.get_dataset_metadata(id)
        dataset = Dataset(metadata)
        dataset.cache_files(data_service)

        return cls(dataset)

    @classmethod
    def from_dataset(cls, dataset: Dataset):
        """Turn Dataset into network component

        Args:
            dataset (obj): Dataset Object.

        Returns:
            obj: network dataset

        """

        return cls(dataset)

    @classmethod
    def from_json_str(cls, json_str, data_service: DataService = None, folder_path=None):
        """Get Dataset from json string.

        Args:
            json_str (str): JSON of the Dataset.
            data_service (obj): Data Service class.
            folder_path (str): File path.

        Returns:
            obj: network dataset

        """
        dataset = Dataset(json.loads(json_str))

        # download file and set local file path using metadata from services
        if data_service is not None:
            dataset.cache_files(data_service)
        # if there is local files associates with the dataset
        elif folder_path is not None:
            dataset.local_file_path = folder_path
        else:
            raise ValueError("You have to either use data services, or given pass local file path.")

        return cls(dataset)

    @classmethod
    def from_files(cls, node_file_path, link_file_path, graph_file_path, network_data_type, link_data_type,
                   node_data_type,
                   graph_data_type):
        """Create Dataset from the file.

        Args:
            node_file_path (str): File path.
            link_file_path (str): File path.
            graph_file_path (str): File path.
            link_data_type (str): Link data type.
            node_data_type (str): Node data type.
            graph_data_type (str): Graph data type.

        Returns:
            obj: Dataset from file.

        """
        metadata = {
            "id": "",
            "dataType": network_data_type,
            "fileDescriptors": [],
            "networkDataset": {
                "link": {
                    "dataType": link_data_type,
                    "fileName": link_file_path
                },
                "node": {
                    "dataType": node_data_type,
                    "fileName": node_file_path
                },
                "graph": {
                    "dataType": graph_data_type,
                    "fileName": graph_file_path
                }
            },
            "format": "shp-network"
        }
        dataset = Dataset(metadata)
        dataset.local_file_path = ""
        return cls(dataset)

    @staticmethod
    def _network_component_from_dataset(dataset: Dataset, network_type="link"):
        """
        Create a dataset object for network from another dataset
        Args:
            dataset: network dataset
            network_type: link, node or graph

        Returns: network component in dataset object

        """
        network_component_filename = dataset.metadata['networkDataset'][network_type]["fileName"]
        network_component_metadata = {
            "dataType": dataset.metadata['networkDataset'][network_type]["dataType"],
            "format": f"shp-{network_type}",
            "id": f"{dataset.id}-{network_type}",
            "fileDescriptors": [fd for fd in dataset.file_descriptors if fd["filename"].find(
                network_component_filename.split(".")[0]) != -1]
        }
        network_component = Dataset(network_component_metadata)
        try:
            file_path = os.path.join(dataset.local_file_path, network_component_filename)
        except FileNotFoundError:
            raise FileNotFoundError("Invalid local file path.")
        network_component.local_file_path = file_path

        return network_component

    def get_nodes(self):
        return self.nodes.get_inventory_reader()

    def get_links(self):
        return self.links.get_inventory_reader()

    def get_graph(self):
        return self.graph.get_csv_reader()

    def get_graph_networkx(self, from_node_fld="fromnode", to_node_fld="tonode", directed=False, numeric=True):
        if directed:
            G = nx.DiGraph()
        else:
            G = nx.Graph()
        for edge in list(self.get_graph()):
            # the csv_reader in the get_graph turns integer into string; sometimes the network needs to be
            # constructed with integer as node id
            if numeric:
                G.add_edge(int(edge[from_node_fld]), int(edge[to_node_fld]))
            else:
                G.add_edge(edge[from_node_fld], edge[to_node_fld])

        return G



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import collections


class NonStructBuildingUtil:
    """Utility methods for the non-structural building damage analysis."""
    BUILDING_FRAGILITY_KEYSBUILDING_FRAGILITY_KEYS = {
        "drift-sensitive fragility id code": ["Drift Sensitive", "DS"],
        "parametric non-retrofit fragility id code": ["Parametric Non-Retrofit", "PNR"],
        "acceleration-sensitive fragility id code": ["Acceleration Sensitive", "AS"],
        "non-retrofit fragility id code": ["as built", "none"]
    }

    DEFAULT_FRAGILITY_KEY_DS = "Drift-Sensitive Fragility ID Code"
    DEFAULT_FRAGILITY_KEY_AS = "Acceleration-Sensitive Fragility ID Code"

    @staticmethod
    def adjust_damage_for_liquefaction(limit_state_probabilities, ground_failure_probabilities):
        """Adjusts building damage probability based on liquefaction ground failure probability
        with the liq_dmg, we know that it is 3 values, the first two are the same.
        The 3rd might be different.
        We always want to apply the first two to all damage states except the highest.

        Args:
            limit_state_probabilities (obj): Limit state probabilities.
            ground_failure_probabilities (list): Ground failure probabilities.

        Returns:
            OrderedDict: Adjusted limit state probability.

        """
        keys = list(limit_state_probabilities.keys())
        adjusted_limit_state_probabilities = collections.OrderedDict()

        for i in range(len(keys)):
            # check and see...if we are trying to use the last ground failure
            # number for something other than the
            # last limit-state-probability, then we should use the
            # second-to-last probability of ground failure instead.

            if i > len(ground_failure_probabilities) - 1:
                prob_ground_failure = ground_failure_probabilities[len(ground_failure_probabilities) - 2]
            else:
                prob_ground_failure = ground_failure_probabilities[i]

            adjusted_limit_state_probabilities[keys[i]] = \
                limit_state_probabilities[keys[i]] + prob_ground_failure \
                - limit_state_probabilities[keys[i]] * prob_ground_failure

        # the final one is the last of limitStates should match with the last of ground failures
        j = len(limit_state_probabilities) - 1
        prob_ground_failure = ground_failure_probabilities[-1]
        adjusted_limit_state_probabilities[keys[j]] = \
            limit_state_probabilities[keys[j]] \
            + prob_ground_failure - limit_state_probabilities[keys[j]] * prob_ground_failure

        return adjusted_limit_state_probabilities

    @staticmethod
    def determine_haz_exposure(hazard_exposure_as, hazard_exposure_ds):
        """
        Determine the hazard exposure of the building based on the
        Args:
            hazard_exposure_as:
            hazard_exposure_ds:

        Returns:

        """
        if hazard_exposure_as == "yes" and hazard_exposure_ds == "yes":
            haz_expose = "yes"
        elif hazard_exposure_as == "error" or hazard_exposure_ds == "error":
            haz_expose = "error"
        elif hazard_exposure_as == "no" and hazard_exposure_ds == "no":
            haz_expose = "no"
        else:
            haz_expose = "partial"

        return haz_expose



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

from urllib.parse import urljoin
from pyincore import IncoreClient
import pyincore.globals as pyglobals
from pyincore.utils import return_http_response
import requests

logger = pyglobals.LOGGER


class SpaceService:
    """Space service client.

    Args:
        client (IncoreClient): Service authentication.

    """

    def __init__(self, client: IncoreClient):
        self.client = client
        self.base_space_url = urljoin(client.service_url, "space/api/spaces/")

    def create_space(self, space_json, timeout=(30, 600), **kwargs):
        """Creates a Space.

        Args:
            space_json (json): A space representation. ID of the Space.
            timeout (tuple): Timeout for the request.
            **kwargs: Additional arguments.

        Returns:
            obj: HTTP response with the created space information.

        """
        url = self.base_space_url
        space_data = {('space', space_json)}
        kwargs["files"] = space_data
        r = self.client.post(url, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def get_spaces(self, dataset_id: str = None, timeout=(30, 600), **kwargs):
        """Retrieve  a Space with the dataset.

        Args:
            dataset_id (str): ID of the Dataset.
            timeout (tuple): Timeout for the request.
            **kwargs: Additional arguments.

        Returns:
            obj: HTTP response with the created space information.

        """
        url = self.base_space_url
        payload = {}
        if dataset_id is not None:
            payload['dataset'] = dataset_id

        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def get_space_by_id(self, space_id: str, timeout=(30, 600), **kwargs):
        """Get space information.

        Args:
            space_id (str): A space representation. ID of the Space.

        Returns:
            obj: HTTP response with the created space information.

        """
        url = urljoin(self.base_space_url, space_id)
        r = self.client.get(url, timeout=timeout, **kwargs)

        return return_http_response(r).json()

    def get_space_by_name(self, space_name: str, timeout=(30, 600), **kwargs):
        """Get space information by querying the name of space.

        Args:
            space_name (str): A space representation. Name of the Space.
            timeout (tuple): Timeout for the request.
            **kwargs: Additional arguments.

        Returns:
            obj: HTTP response with the returned space information.

        """
        r = self.client.get(self.base_space_url, params={"name": space_name}, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def update_space(self, space_id: str, space_json, timeout=(30, 600), **kwargs):
        """Updates a Space.

        Args:
            space_id (str): ID of the space to update.
            space_json (json): JSON representing a space update.
            timeout (tuple): Timeout for the request.
            **kwargs: Additional arguments.

        Returns:
            obj: HTTP response with the updated space.

        """
        url = urljoin(self.base_space_url, space_id)
        space_data = {('space', space_json)}
        kwargs["files"] = space_data
        r = self.client.put(url, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def add_to_space_by_name(self, space_name: str, dataset_id: str):
        """Add dataset to a space by using space name and dataset id.

        Args:
            space_name (str): Name of the space that dataset id will go to.
            dataset_id (json): ID of the dataset that will be copied in to the space.

        Returns:
            obj: HTTP response with the updated space.

        """
        space_id = self.get_space_by_name(space_name)[0]

        response = self.add_dataset_to_space(space_id["id"], dataset_id)

        return response

    def remove_from_space_by_name(self, space_name: str, dataset_id: str):
        """Remove dataset from a space by using space name and dataset id.

        Args:
            space_name (str): Name of the space that dataset id will go to.
            dataset_id (json): ID of the dataset that will be copied in to the space.

        Returns:
            obj: HTTP response with the updated space.

        """
        space_id = self.get_space_by_name(space_name)[0]

        response = self.remove_dataset_from_space(space_id["id"], dataset_id)

        return response

    def remove_dataset_from_space(self, space_id: str, dataset_id: str, timeout=(30, 600), **kwargs):
        """Remove dataset from the space using dataset id and space id

        Args:
            space_id (str): ID of the space to update.
            dataset_id (str): ID of the dataset to be removed from the space.
            timeout (tuple): Timeout for the request.
            **kwargs: Additional arguments.

        Returns:
            obj: HTTP response with the updated space.

        """
        url = urljoin(self.base_space_url, space_id + "/members/" + dataset_id)

        r = self.client.delete(url, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def add_dataset_to_space(self, space_id: str, dataset_id: str, timeout=(30, 600), **kwargs):
        """Add member to a Space.

        Args:
            space_id (str): ID of the space to update.
            dataset_id (str): ID of the dataset to be added to the space.
            timeout (tuple): Timeout for the request.
            **kwargs: Additional arguments.

        Returns:
            obj: HTTP response with the updated space.

        """
        url = urljoin(self.base_space_url, space_id + "/members/" + dataset_id)

        r = self.client.post(url, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def grant_privileges_to_space(self, space_id: str, privileges_json, timeout=(30, 600), **kwargs):
        """Updates a Space.

        Args:
            space_id (str): ID of the space to update.
            privileges_json (json): JSON representing a space update.
            timeout (tuple): Timeout for the request.
            **kwargs: Additional arguments.

        Returns:
            obj: HTTP response with space privileges.

        """
        url = urljoin(self.base_space_url, space_id + "/grant")
        space_privileges = {('grant', privileges_json)}
        kwargs["files"] = space_privileges
        r = self.client.post(url, timeout=timeout, **kwargs)

        return return_http_response(r).json()



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/
import collections
import csv
import math
import os
import re
import numpy as np
from typing import List, Dict
from collections import Counter

from decimal import getcontext, Decimal

from pyincore import DataService
from pyincore.globals import DAMAGE_PRECISION
from pyincore.utils import evaluateexpression


class AnalysisUtil:
    """Utility methods for analysis"""
    DOCSTR_FORMAT = "$DESC$ \n\n" \
                    "Args: \n\t" \
                    "$ARGS$ " \
                    "\n" \
                    "Returns: \n\t" \
                    "$RETS$ "

    getcontext().prec = DAMAGE_PRECISION

    @staticmethod
    def update_precision(num, precision: int = DAMAGE_PRECISION):
        try:
            r = round(float(num), precision)
            return r
        except TypeError:
            print("Error trying to round non numeric value")
            raise

    @staticmethod
    def update_precision_of_dicts(states: dict) -> dict:
        updated_states = {key: AnalysisUtil.update_precision(states[key]) for key in states}
        return updated_states

    @staticmethod
    def update_precision_of_lists(hazard_vals: List) -> List:
        updated_hazard_vals = []
        for val in hazard_vals:
            if val is not None:
                if math.ceil(val) == -9999:  # if it's an error code(-9999.x) do not update precision
                    updated_hazard_vals.append(val)
                else:
                    updated_hazard_vals.append(AnalysisUtil.update_precision(val))
            else:
                updated_hazard_vals.append(None)
        return updated_hazard_vals

    @staticmethod
    def float_to_decimal(num: float):

        # Helper function to check if a string is a float
        def is_float(string):
            try:
                float(string)
                return True
            except ValueError:
                return False

        if is_float(num):
            return Decimal(str(num))
        else:
            return np.nan

    @staticmethod
    def float_list_to_decimal(num_list: list):
        return [Decimal(str(num)) for num in num_list]

    @staticmethod
    def float_dict_to_decimal(num_dict: dict):
        return {key: Decimal(str(num_dict[key])) for key in num_dict}

    @staticmethod
    def dmg_string_dict_to_dmg_float_dict(dmg_dict: dict):
        float_dmg_dict = {}
        for key in dmg_dict:
            if key != 'guid' and key != 'haz_expose':
                if dmg_dict[key] == '':
                    float_dmg_dict[key] = np.nan
                else:
                    float_dmg_dict[key] = float(dmg_dict[key])
            else:
                if dmg_dict[key] != '':
                    float_dmg_dict[key] = dmg_dict[key]
                else:
                    float_dmg_dict[key] = np.nan
        return float_dmg_dict

    @staticmethod
    def calculate_mean_damage(dmg_ratio_tbl, dmg_intervals,
                              damage_interval_keys, is_bridge=False,
                              bridge_spans=1):
        if len(damage_interval_keys) < 4:
            raise ValueError("we only accept 4 damage or more than 4 interval keys!")

        float_dmg_intervals = AnalysisUtil.dmg_string_dict_to_dmg_float_dict(dmg_intervals)

        output = collections.OrderedDict()
        if len(dmg_ratio_tbl) == 5:
            output['meandamage'] = float(
                float(dmg_ratio_tbl[1]["Best Mean Damage Ratio"]) * float_dmg_intervals[damage_interval_keys[0]] + \
                float(dmg_ratio_tbl[2]["Best Mean Damage Ratio"]) * float_dmg_intervals[damage_interval_keys[1]] + \
                float(dmg_ratio_tbl[3]["Best Mean Damage Ratio"]) * float_dmg_intervals[damage_interval_keys[2]] +\
                float(dmg_ratio_tbl[4]["Best Mean Damage Ratio"]) * float_dmg_intervals[damage_interval_keys[3]])

        elif len(dmg_ratio_tbl) == 4:
            output['meandamage'] = float(
                float(dmg_ratio_tbl[0]["Mean Damage Factor"]) * float_dmg_intervals[damage_interval_keys[0]] + \
                float(dmg_ratio_tbl[1]["Mean Damage Factor"]) * float_dmg_intervals[damage_interval_keys[1]] + \
                float(dmg_ratio_tbl[2]["Mean Damage Factor"]) * float_dmg_intervals[damage_interval_keys[2]] + \
                float(dmg_ratio_tbl[3]["Mean Damage Factor"]) * float_dmg_intervals[damage_interval_keys[3]])

        elif len(dmg_ratio_tbl) == 6 and is_bridge:
            # this is for bridge
            weight_slight = float(dmg_ratio_tbl[1]['Best Mean Damage Ratio'])
            weight_moderate = float(dmg_ratio_tbl[2]['Best Mean Damage Ratio'])
            weight_extensive = float(
                dmg_ratio_tbl[3]['Best Mean Damage Ratio'])
            weight_collapse0 = float(
                dmg_ratio_tbl[4]['Best Mean Damage Ratio'])
            weight_collapse1 = float(
                dmg_ratio_tbl[5]['Best Mean Damage Ratio'])

            output['meandamage'] = \
                weight_slight * float_dmg_intervals[damage_interval_keys[1]] + \
                weight_moderate * float_dmg_intervals[damage_interval_keys[2]] + \
                weight_extensive * float_dmg_intervals[damage_interval_keys[3]]

            if bridge_spans >= 3:
                output[
                    'meandamage'] += weight_collapse1 / bridge_spans *float_dmg_intervals[damage_interval_keys[4]]
            else:
                output['meandamage'] += weight_collapse0 * float_dmg_intervals[damage_interval_keys[4]]
        else:
            raise ValueError('We cannot handle this damage ratio format.')

        return output

    @staticmethod
    def calculate_mean_damage_std_deviation(dmg_ratio_tbl, dmg,
                                            mean_damage, damage_interval_keys):

        float_dmg = AnalysisUtil.dmg_string_dict_to_dmg_float_dict(dmg)
        output = collections.OrderedDict()
        result = 0.0
        idx = 0
        for key in damage_interval_keys:
            result += float_dmg[key] * (math.pow(
                float(dmg_ratio_tbl[idx]["Mean Damage Factor"]), 2) + math.pow(
                float(dmg_ratio_tbl[idx]["Deviation Damage Factor"]), 2))
            idx += 1

        output['mdamagedev'] = math.sqrt(result - math.pow(mean_damage, 2))
        return output

    @staticmethod
    def get_expected_damage(mean_damage, dmg_ratios):
        """Calculates mean damage.

        Args:
            mean_damage (float): Mean damage value.
            dmg_ratios (obj): Damage ratios, descriptions and states.

        Returns:
            float: A value of the damage state.

        """
        no_dmg_bound = [float(dmg_ratios[1]["Lower Bound"]),
                        float(dmg_ratios[1]["Upper Bound"])]
        slight_bound = [float(dmg_ratios[2]["Lower Bound"]),
                        float(dmg_ratios[2]["Upper Bound"])]
        moderate_bound = [float(dmg_ratios[3]["Lower Bound"]),
                          float(dmg_ratios[3]["Upper Bound"])]
        extensive_bound = [float(dmg_ratios[4]["Lower Bound"]),
                           float(dmg_ratios[4]["Upper Bound"])]
        collapse_bound = [float(dmg_ratios[5]["Lower Bound"]),
                          float(dmg_ratios[5]["Upper Bound"])]
        if no_dmg_bound[0] <= mean_damage <= no_dmg_bound[1]:
            idx = 1
        elif slight_bound[0] <= mean_damage <= slight_bound[1]:
            idx = 2
        elif moderate_bound[0] <= mean_damage <= moderate_bound[1]:
            idx = 3
        elif extensive_bound[0] <= mean_damage <= extensive_bound[1]:
            idx = 4
        elif collapse_bound[0] <= mean_damage <= collapse_bound[1]:
            idx = 5
        else:
            idx = 1
        return dmg_ratios[idx]["Damage State"]

    @staticmethod
    def determine_parallelism_locally(self, number_of_loops,
                                      user_defined_parallelism=0):
        """Determine the parallelism on the current compute node.

        Args:
            number_of_loops: total number of loops will be executed on current compute node
            user_defined_parallelism: a customized parallelism specified by users

        Returns:
            int: parallelism on current compute node

        """

        # gets the local cpu number
        number_of_cpu = os.cpu_count()
        if number_of_loops > 0:
            if user_defined_parallelism > 0:
                return min(number_of_cpu, number_of_loops,
                           user_defined_parallelism)
            else:
                return min(number_of_cpu, number_of_loops)
        else:
            return number_of_cpu

    @staticmethod
    def create_result_dataset(datasvc: DataService, parentid: str,
                              result_files: List[str], title: str,
                              output_metadata: Dict[str, str]):
        # Result metadata
        properties = output_metadata
        properties["title"] = title
        properties["sourceDataset"] = parentid

        # Create child dataset with parent dataset as sourceDataset
        result_dataset = datasvc.create_dataset(properties)
        result_dataset_id = result_dataset["id"]

        # Attach files to result dataset
        datasvc.add_files_to_dataset(result_dataset_id, result_files)

        return result_dataset_id

    @staticmethod
    def adjust_damage_for_liquefaction(limit_state_probabilities, ground_failure_probabilities):
        """Adjusts building damage probability based on liquefaction ground failure probability
        with the liq_dmg, we know that it is 3 values, the first two are the same.
        The 3rd might be different.
        We always want to apply the first two to all damage states except the highest.

        Args:
            limit_state_probabilities (obj): Limit state probabilities.
            ground_failure_probabilities (list): Ground failure probabilities.

        Returns:
            OrderedDict: Adjusted limit state probability.

        """
        keys = list(limit_state_probabilities.keys())
        adjusted_limit_state_probabilities = collections.OrderedDict()

        for i in range(len(keys)):
            # check and see...if we are trying to use the last ground failure
            # number for something other than the
            # last limit-state-probability, then we should use the
            # second-to-last probability of ground failure instead.

            if i > len(ground_failure_probabilities) - 1:
                prob_ground_failure = ground_failure_probabilities[len(ground_failure_probabilities) - 2]
            else:
                prob_ground_failure = ground_failure_probabilities[i]

            adjusted_limit_state_probabilities[keys[i]] = \
                limit_state_probabilities[keys[i]] + prob_ground_failure \
                - limit_state_probabilities[keys[i]] * prob_ground_failure

        # the final one is the last of limitStates should match with the last of ground failures
        j = len(limit_state_probabilities) - 1
        prob_ground_failure = ground_failure_probabilities[-1]
        adjusted_limit_state_probabilities[keys[j]] = \
            limit_state_probabilities[keys[j]] \
            + prob_ground_failure - limit_state_probabilities[keys[j]] * prob_ground_failure

        return adjusted_limit_state_probabilities

    @staticmethod
    def adjust_limit_states_for_pgd(limit_states, pgd_limit_states):
        try:
            adj_limit_states = collections.OrderedDict()

            for key, value in limit_states.items():
                adj_limit_states[key] = limit_states[key] + pgd_limit_states[key] - \
                                        (limit_states[key] * pgd_limit_states[key])

            return AnalysisUtil.update_precision_of_dicts(adj_limit_states)

        except KeyError as e:
            print('Mismatched keys encountered in the limit states')
            print(str(e))

    @staticmethod
    def get_csv_table_rows(csv_reader: csv.DictReader, ignore_first_row=True):
        csv_rows = []

        row_index = 0
        for row in csv_reader:
            if ignore_first_row:
                # Ignore the first row
                if row_index > 0:
                    csv_rows.append(row)
            else:
                csv_rows.append(row)

            row_index += 1

        return csv_rows

    @staticmethod
    def create_gdocstr_from_spec(specs):
        """

        Args:
            specs (dict): Json of the specs for each analysis

        Returns:
            str: Google format docstrings to copy for the run() method of any analysis

        """
        desc = specs['description']
        args = ""
        rets = ""

        for dataset in specs['input_datasets']:
            is_opt = ""
            if not dataset['required']:
                is_opt = ", " + "optional"

            args = \
                args + dataset['id'] + "(str" + is_opt + ") : " \
                + dataset['description'] + ". " \
                + AnalysisUtil.get_custom_types_str(dataset['type']) + "\n\t"

        for param in specs['input_parameters']:
            is_opt = ""
            if not param['required']:
                is_opt = ", " + "optional"

            args = \
                args + param['id'] + "(" + AnalysisUtil.get_type_str(param['type']) + is_opt + ") : " \
                + param['description'] + "\n\t"

        for dataset in specs['output_datasets']:
            rets = rets + dataset['id'] + ": " \
                   + dataset[
                       'description'] + ". " + AnalysisUtil.get_custom_types_str(
                dataset['type']) + "\n\t"

        docstr = AnalysisUtil.DOCSTR_FORMAT.replace("$DESC$", desc).replace(
            "$ARGS$",
            args).replace("$RETS$", rets)

        print(docstr)

    @staticmethod
    def get_type_str(class_type):
        """

        Args:
            class_type (str): Example: <class 'int'>

        Returns:
            str: Text inside first single quotes of a string

        """

        t = str(class_type)
        match = re.search('\'([^"]*)\'', t)
        if match is not None:
            return match.group(1)
        return None

    @staticmethod
    def get_custom_types_str(types):
        """

        Args:
            types (str, list): Can be string or List of strings

        Returns:
            str: Formatted string with applicable datatypes used to generate docstrigns from specs

        """
        custom_types_str = 'Applicable dataset type(s): '
        if (isinstance(types, str)):
            return custom_types_str + types
        if (isinstance(types, list)):
            if (len(types) > 1):
                idx = 0
                for type in types:
                    if idx < len(types) - 1:
                        custom_types_str += type + ", "
                    else:
                        custom_types_str += type
                    idx = idx + 1
                return custom_types_str
            else:
                return types[0]

    @staticmethod
    def chunks(lst, n):
        """Yield successive n-sized chunks from lst."""
        for i in range(0, len(lst), n):
            yield lst[i:i + n]

    @staticmethod
    def get_hazard_demand_type(building, fragility_set, hazard_type):
        """
        Get hazard demand type. This method is intended to replace get_hazard_demand_type. Fragility_set is not a
        json but a fragilityCurveSet object now.

        Args:
            building (obj): A JSON mapping of a geometric object from the inventory: current building.
            fragility_set (obj): FragilityCurveSet object
            hazard_type (str): A hazard type such as earthquake, tsunami etc.

        Returns:
            str: A hazard demand type.

        """
        BLDG_STORIES = "no_stories"
        PROPERTIES = "properties"
        BLDG_PERIOD = "period"

        fragility_hazard_type = fragility_set.demand_types[0].lower()
        hazard_demand_type = fragility_hazard_type

        if hazard_type.lower() == "earthquake":
            num_stories = building[PROPERTIES][BLDG_STORIES]
            # Get building period from the fragility if possible

            building_args = fragility_set.construct_expression_args_from_inventory(building)
            building_period = fragility_set.fragility_curves[0].get_building_period(
                fragility_set.curve_parameters, **building_args)

            if fragility_hazard_type.endswith('sa') and fragility_hazard_type != 'sa':
                # This fixes a bug where demand type is in a format similar to 1.0 Sec Sa
                if len(fragility_hazard_type.split()) > 2:
                    building_period = fragility_hazard_type.split()[0]
                    fragility_hazard_type = "Sa"

            hazard_demand_type = fragility_hazard_type

            # This handles the case where some fragilities only specify Sa, others a specific period of Sa
            if not hazard_demand_type.endswith('pga'):
                # If the fragility does not contain the period calculation, check if the dataset has it
                if building_period == 0.0 and BLDG_PERIOD in building[PROPERTIES]:
                    if building[PROPERTIES][BLDG_PERIOD] > 0.0:
                        building_period = building[PROPERTIES][BLDG_PERIOD]

                hazard_demand_type = str(building_period) + " " + fragility_hazard_type
        elif hazard_type.lower() == "tsunami":
            if hazard_demand_type == "momentumflux":
                hazard_demand_type = "mmax"
            elif hazard_demand_type == "inundationdepth":
                hazard_demand_type = "hmax"

        return hazard_demand_type

    @staticmethod
    def get_hazard_demand_types_units(building, fragility_set, hazard_type, allowed_demand_types):
        """
        Get hazard demand type. This method is intended to replace get_hazard_demand_type. Fragility_set is not a
        json but a fragilityCurveSet object now.

        Args:
            building (obj): A JSON mapping of a geometric object from the inventory: current building.
            fragility_set (obj): FragilityCurveSet object
            hazard_type (str): A hazard type such as earthquake, tsunami etc.
            allowed_demand_types (list): A list of allowed demand types in lowercase

        Returns:
            str: A hazard demand type.

        """
        BLDG_STORIES = "no_stories"
        PROPERTIES = "properties"
        BLDG_PERIOD = "period"

        fragility_demand_types = fragility_set.demand_types
        fragility_demand_units = fragility_set.demand_units
        adjusted_demand_types = []
        adjusted_demand_units = []
        adjusted_to_original = {}

        for index, demand_type in enumerate(fragility_demand_types):
            original_demand_type = demand_type
            adjusted_to_original[original_demand_type] = original_demand_type

            demand_type = demand_type.lower()
            adjusted_demand_type = demand_type
            adjusted_demand_unit = fragility_demand_units[index]
            # TODO: Due to the mismatch in demand types names on DFR3 vs hazard service, we should refactor this before
            # TODO: using expression based fragilities for tsunami & earthquake
            if hazard_type.lower() == "tsunami":
                if demand_type == "momentumflux":
                    adjusted_demand_type = "mmax"
                elif demand_type == "inundationdepth":
                    adjusted_demand_type = "hmax"
                if adjusted_demand_type not in allowed_demand_types:
                    continue
            elif hazard_type.lower() == "earthquake":
                # TODO temp fix
                allowed = False
                for allowed_demand_type in allowed_demand_types:
                    if allowed_demand_type in demand_type:
                        allowed = True
                        break
                if allowed:
                    num_stories = building[PROPERTIES][BLDG_STORIES]
                    # Get building period from the fragility if possible
                    building_args = fragility_set.construct_expression_args_from_inventory(building)
                    building_period = fragility_set.fragility_curves[0].get_building_period(
                        fragility_set.curve_parameters, **building_args)

                    # TODO: There might be a bug here as this is not handling SD
                    if demand_type.endswith('sa'):
                        # This fixes a bug where demand type is in a format similar to 1.0 Sec Sa
                        if demand_type != 'sa':
                            if len(demand_type.split()) > 2:
                                building_period = demand_type.split()[0]
                                adjusted_demand_type = str(building_period) + " " + "SA"
                        else:
                            if building_period == 0.0:
                                if BLDG_PERIOD in building[PROPERTIES] and building[PROPERTIES][BLDG_PERIOD] > 0.0:

                                    building_period = building[PROPERTIES][BLDG_PERIOD]
                                else:
                                    # try to calculate the period from the expression
                                    for param in fragility_set.curve_parameters:
                                        if param["name"].lower() == "period":
                                            # TODO: This is a hack and expects a parameter with name "period" present.
                                            # This can potentially cause naming conflicts in some fragilities

                                            building_period = evaluateexpression.evaluate(param["expression"],
                                                                                          {"num_stories": num_stories})
                                            # TODO: num_stories logic is not tested. should find a fragility with
                                            # periodEqnType = 2 or 3 to test. periodEqnType = 1 doesn't need
                                            # num_stories.

                            adjusted_demand_type = str(building_period) + " " + "SA"
                else:
                    continue
            else:
                if demand_type not in allowed_demand_types:
                    continue
            adjusted_to_original[adjusted_demand_type] = original_demand_type
            adjusted_demand_types.append(adjusted_demand_type)
            adjusted_demand_units.append(adjusted_demand_unit)
        return adjusted_demand_types, adjusted_demand_units, adjusted_to_original

    @staticmethod
    def group_by_demand_type(inventories, fragility_sets, hazard_type="earthquake", is_building=False):
        """
        This method should replace group_by_demand_type in the future. Fragility_sets is not list of dictionary (
        json) anymore but a list of FragilityCurveSet objects

        Args:
            inventories (dict): dictionary of {id: intentory}
            fragility_sets (dict): fragility_sets
            hazard_type (str): default to earthquake
            is_building (bool): if the inventory is building or not

        Returns:
            dict: dGrouped inventory with { (demandunit, demandtype):[inventory ids] }

        """
        grouped_inventory = dict()
        for fragility_id, frag in fragility_sets.items():

            # TODO this method will be deprecated so this is temporary fix
            demand_type = frag.demand_types[0]
            demand_unit = frag.demand_units[0]

            if is_building:
                inventory = inventories[fragility_id]
                demand_type = AnalysisUtil.get_hazard_demand_type(inventory, frag, hazard_type)

            tpl = (demand_type, demand_unit)
            grouped_inventory.setdefault(tpl, []).append(fragility_id)

        return grouped_inventory

    @staticmethod
    def get_exposure_from_hazard_values(hazard_vals, hazard_type):
        """Finds if a point is exposed to hazard based on all the demand values present.
        Returns \"n/a\" for earthquake, tsunami, hurricane and hurricane windfields

        Args:
            hazard_vals (list): List of hazard values returned by the service for a particular point
            hazard_type (str): Type of the hazard

        Returns:
            str: If hazard is exposed or not. Can be one of 'yes', 'no', 'partial', 'error' or 'n/a'

        """

        # This method should handle other 'error' cases when service is able to handle exceptions by returning -9999/NaN
        # and return "error". hazard_type parameter should be removed when all hazards are supported on the service.

        if len(hazard_vals) == 0:
            return "error"

        supported_hazards = ["earthquake", "tsunami", "tornado", "hurricane", "flood"]
        hazard_exposure = "n/a"
        if hazard_type.lower() in supported_hazards:
            if AnalysisUtil.do_hazard_values_have_errors(hazard_vals):
                hazard_exposure = "error"
            else:
                cnt_hazard_vals = Counter(hazard_vals)
                if None in cnt_hazard_vals:
                    if cnt_hazard_vals.get(None) == len(hazard_vals):
                        hazard_exposure = "no"
                    else:
                        hazard_exposure = "partial"
                else:
                    hazard_exposure = "yes"  # none of the values are nulls

        return hazard_exposure

    @staticmethod
    def do_hazard_values_have_errors(hazard_vals):
        """ Checks if any of the hazard values have errors

        Args:
            hazard_vals(list): List of hazard values returned by the service for a particular point

        Returns: True if any of the values are error codes such as -9999.1, -9999.2 etc.

        """
        return any("-9999" in str(val) for val in hazard_vals)

    @staticmethod
    def get_discretized_restoration(restoration_curve_set, discretized_days):
        """Converts discretized restoration times into a dictionary from continuous curve

        Args:
            restoration_curve_set(obj):
            discretized_days(list):
        Returns:
            dict: discretized restoration for each day {day1: [100, 50, 9, 4, 3], day3: [100,
            100, 50, 13, 4], etc }
        """

        class_restoration = {}
        for time in discretized_days:
            restoration_times = restoration_curve_set.calculate_restoration_rates(time=time)
            # Key (e.g. day1, day3)
            time_key = "day"+str(time)

            restoration = [1, restoration_times['PF_0'], restoration_times['PF_1'], restoration_times['PF_2'],
                           restoration_times['PF_3']]

            class_restoration[time_key] = restoration

        return class_restoration



# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import numpy as np
import pandas as pd
import warnings
from pyincore import BaseAnalysis
from pyincore.analyses.populationdislocation.populationdislocationutil import PopulationDislocationUtil


class PopulationDislocation(BaseAnalysis):
    """Population Dislocation Analysis computes dislocation for each residential structure based on the direct
    economic damage. The dislocation is calculated from four probabilities of dislocation based on a random normal
    distribution of the four damage factors presented by Bai, Hueste, Gardoni 2009.

    These four damage factors correspond to value loss. The sum of the four probabilities multiplied
    by the four probabilities of damage states was used as the probability for dislocation.

    This is different from Lin 2008
    http://hrrc.arch.tamu.edu/publications/research%20reports/08-05R%20Dislocation%20Algorithm%203.pdf
    which calculates a value loss which is the sum of the four damage factors times the four probabilities
    of damage. The two approaches produce different results.

    Args:
        incore_client (IncoreClient): Service authentication.

    """

    def __init__(self, incore_client):
        super(PopulationDislocation, self).__init__(incore_client)

    def get_spec(self):
        return {
            'name': 'population-dislocation',
            'description': 'Population Dislocation Analysis',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'Result CSV dataset name',
                    'type': str
                },
                {
                    'id': 'seed',
                    'required': True,
                    'description': 'Seed to ensure replication if run as part of a probabilistic analysis, '
                                   'for example in connection with Housing Unit Allocation analysis.',
                    'type': int
                }
            ],
            'input_datasets': [
                {
                    'id': 'building_dmg',
                    'required': True,
                    'description': 'Building damage results CSV file',
                    'type': ['ergo:buildingInventoryVer4',
                             'ergo:buildingDamageVer5',
                             'ergo:buildingDamageVer6',
                             'ergo:buildingInventory']
                },
                {
                    'id': 'housing_unit_allocation',
                    'required': True,
                    'description': 'A csv file with the merged dataset of the inputs, aka Probabilistic'
                                   'House Unit Allocation',
                    'type': ['incore:housingUnitAllocation']
                },
                {
                    'id': 'block_group_data',
                    'required': True,
                    'description': 'Block group racial distribution census CSV data',
                    'type': ['incore:blockGroupData']
                },
                {
                    'id': 'value_loss_param',
                    'required': True,
                    'description': 'A table with value loss beta distribution parameters based on Bai et al. 2009',
                    'type': ['incore:valuLossParam']
                }
            ],
            'output_datasets': [
                {
                    'id': 'result',
                    'parent_type': 'population_block',
                    'description': 'A csv file with population dislocation result '
                                   'aggregated to the block group level',
                    'type': 'incore:popDislocation'
                }
            ]
        }

    def run(self):
        """Executes the Population dislocation analysis.

        Returns:
            bool: True if successful, False otherwise.

        """
        # Get seed
        seed_i = self.get_parameter("seed")

        result_name = self.get_parameter("result_name")

        # Building damage dataset
        building_dmg = self.get_input_dataset("building_dmg").get_dataframe_from_csv(low_memory=False)

        # Housing unit allocation dataset
        housing_unit_alloc = self.get_input_dataset("housing_unit_allocation"). \
            get_dataframe_from_csv(low_memory=False)

        # Block group dataset
        bg_data = self.get_input_dataset("block_group_data").get_dataframe_from_csv(low_memory=False)

        # Get value loss parameters
        value_loss = self.get_input_dataset("value_loss_param").get_dataframe_from_csv(low_memory=False)
        value_loss.set_index('damagestate', inplace=True)

        merged_block_inv = PopulationDislocationUtil.merge_damage_housing_block(
            building_dmg, housing_unit_alloc, bg_data
        )

        # Returns dataframe
        merged_final_inv = self.get_dislocation(seed_i, merged_block_inv, value_loss)
        csv_source = "dataframe"
        self.set_result_csv_data("result", merged_final_inv, result_name, "dataframe")

        return True

    def get_dislocation(self, seed_i: int, inventory: pd.DataFrame, value_loss: pd.DataFrame):
        """Calculates dislocation probability.

        Probability of dislocation, a binary variable based on the logistic probability of dislocation.
        A random number between 0 and 1 was assigned to each household.
        If the random number was less than the probability of dislocation
        then the household was determined to dislocate. This follows the logic
        that households with a greater chance of dislocated were more likely
        to have a random number less than the probability predicted.

        Args:
            seed_i (int): Seed for random number generator to ensure replication if run as part
            of a stochastic analysis, for example in connection with housing unit allocation analysis.
            inventory (pd.DataFrame): Merged building, housing unit allocation and block group inventories
            value_loss (pd.DataFrame): Table used for value loss estimates, beta distribution

        Returns:
            pd.DataFrame: An inventory with probabilities of dislocation in a separate column

        """
        # pd.Series to np.array
        # creats d_sf column it if it does not exist, overwrites d_sf values if it does
        if "huestimate" in inventory.columns:
            inventory["d_sf"] = (inventory["huestimate"] == 1).astype(int)
        elif "huestimate_x" in inventory.columns:
            inventory = PopulationDislocationUtil.compare_columns(inventory,
                                                                  "huestimate_x",
                                                                  "huestimate_y", True)
            if "huestimate_x-huestimate_y" in inventory.columns:
                exit("Column huestimate is ambiguous, check the input datasets!")
            else:
                inventory["d_sf"] = (inventory["huestimate"] == 1).astype(int)

        # drop d_sf_x, d_sf_y if they exist
        if "d_sf_x" in inventory.columns:
            inventory = inventory.drop(columns=["d_sf_x"])
        if "d_sf_y" in inventory.columns:
            inventory = inventory.drop(columns=["d_sf_y"])
        dsf = inventory["d_sf"].values
        hue = inventory["huestimate"].values.astype(int)
        pbd = inventory["pblackbg"].values
        phd = inventory["phispbg"].values

        prob0 = inventory["DS_0"].values
        prob1 = inventory["DS_1"].values
        prob2 = inventory["DS_2"].values
        prob3 = inventory["DS_3"].values

        # include random value loss by damage state
        rploss0 = PopulationDislocationUtil.get_random_loss(seed_i, value_loss, "DS_0", dsf.size)
        rploss1 = PopulationDislocationUtil.get_random_loss(seed_i, value_loss, "DS_1", dsf.size)
        rploss2 = PopulationDislocationUtil.get_random_loss(seed_i, value_loss, "DS_2", dsf.size)
        rploss3 = PopulationDislocationUtil.get_random_loss(seed_i, value_loss, "DS_3", dsf.size)

        inventory["rploss_0"] = rploss0
        inventory["rploss_1"] = rploss1
        inventory["rploss_2"] = rploss2
        inventory["rploss_3"] = rploss3

        prob0_disl = PopulationDislocationUtil.get_disl_probability(rploss0, dsf, pbd, phd)
        prob1_disl = PopulationDislocationUtil.get_disl_probability(rploss1, dsf, pbd, phd)
        prob2_disl = PopulationDislocationUtil.get_disl_probability(rploss2, dsf, pbd, phd)
        prob3_disl = PopulationDislocationUtil.get_disl_probability(rploss3, dsf, pbd, phd)

        #  dislocation probability is 0 if the damage is set to 100% probability (insignificant, DS_0 = 1).
        #  DS_0 bulding does not distinguish between in and out hazard boundaries. All DS_0 = 1 are set to
        #  zero dislocation probability.
        prob0_disl = np.where(prob0 == 1, 0, prob0_disl)

        # total_prob_disl is the sum of the probability of dislocation at four damage states
        # times the probability of being in that damage state.
        total_prob_disl = prob0_disl * prob0 + prob1_disl * prob1 + prob2_disl * prob2 + prob3_disl * prob3

        inventory["prdis"] = total_prob_disl

        # Randomly assign dislocation based on probability of dislocation
        random_generator = np.random.RandomState(seed_i)
        randomdis = random_generator.uniform(0, 1, total_prob_disl.size)

        # Probability of dislocation, a binary variable based on the logistic probability of dislocation.
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        # comparisons involving np.nan outputs False, supress RuntimeWarning above
        dislocated = np.less_equal(randomdis, total_prob_disl)

        inventory["dislocated"] = dislocated

        return inventory



# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import copy
import networkx as nx
import random
from pyincore.analyses.transportationrecovery.nsga2 import Solution
from pyincore.analyses.transportationrecovery import WIPW as WIPW
from pyincore.analyses.transportationrecovery.transportationrecoveryutil import TransportationRecoveryUtil


class PostDisasterLongTermSolution(Solution):
    """
    Solution for the post disaster long term recovery function.
    """

    # for repair time of bridge assignment
    # slight damage state
    slightRepair = 0.6
    # moderate damage state
    modRepair = 2.5
    # extensive damage state
    extRepair = 75
    # complete damage state
    compRepair = 230

    def __init__(self, candidates, node_df, arc_df, bridge_df, bridge_damage_value,
                 network, pm, all_ipw, path_adt):
        """
        initialize the chromosomes
        """
        Solution.__init__(self, 2)
        self.candidates = candidates
        self.node_df = node_df
        self.arc_df = arc_df
        self.bridge_df = bridge_df
        self.attributes = [i for i in range(len(self.candidates))]
        self.bridge_damage_value = bridge_damage_value
        self.network = network
        self.pm = pm
        self.all_ipw = all_ipw
        self.path_adt = path_adt

        # random sort the sequence
        random.shuffle(self.attributes)

    def evaluate_solution(self, final):
        """
        Implementation of evaluation for all solutions
        """
        temp_bridge_damage_value = copy.deepcopy(self.bridge_damage_value)

        # input the total number of crew groups repairing the bridge at the
        # same time
        simax = 8

        # determine the bridge schedule sequence for each candidate
        # (Chromosome)
        candidate_schedule = {}
        for i in range(len(self.candidates)):
            candidate_schedule[self.attributes[i]] = self.candidates[i]

        trt = 0
        srt = 0

        start = {}
        end = {}
        schedule_time = []

        l = copy.deepcopy(self.attributes)

        for i in range(len(l)):
            if i <= simax - 1:

                # repair start from time 0
                start[candidate_schedule[i]] = 0.0

                # if damage state of bridge is slight damage, repair time
                # is slightRepair
                if temp_bridge_damage_value[candidate_schedule[i]] == 1:
                    end[candidate_schedule[i]] \
                        = start[candidate_schedule[i]] \
                          + PostDisasterLongTermSolution.slightRepair

                # if damage state of bridge is moderate damage, repair time
                # is modRepair
                elif temp_bridge_damage_value[candidate_schedule[i]] == 2:
                    end[candidate_schedule[i]] \
                        = start[candidate_schedule[i]] \
                          + PostDisasterLongTermSolution.modRepair

                # if damage state of bridge is extensive damage, repair time
                # is extRepair
                elif temp_bridge_damage_value[candidate_schedule[i]] == 3:
                    end[candidate_schedule[i]] \
                        = start[candidate_schedule[i]] \
                          + PostDisasterLongTermSolution.extRepair

                # if damage state of bridge is complete damage, repair time
                # is compRepair
                else:
                    end[candidate_schedule[i]] \
                        = start[candidate_schedule[i]] \
                          + PostDisasterLongTermSolution.compRepair

                # store the ending time
                schedule_time.append(end[candidate_schedule[i]])

                # sort the complete time for scheduled bridges to hold maximum
                # simultaneously intervention
                schedule_time.sort()

            else:
                # the repair time of next bridge will be start from the minimum
                # end time of pre-scheduled bridges
                start[candidate_schedule[i]] = schedule_time.pop(0)

                if temp_bridge_damage_value[candidate_schedule[i]] == 1:
                    end[candidate_schedule[i]] \
                        = start[candidate_schedule[i]] \
                          + PostDisasterLongTermSolution.slightRepair

                elif temp_bridge_damage_value[candidate_schedule[i]] == 2:
                    end[candidate_schedule[i]] \
                        = start[candidate_schedule[i]] \
                          + PostDisasterLongTermSolution.modRepair

                elif temp_bridge_damage_value[candidate_schedule[i]] == 3:
                    end[candidate_schedule[i]] \
                        = start[candidate_schedule[i]] \
                          + PostDisasterLongTermSolution.extRepair

                else:
                    end[candidate_schedule[i]] \
                        = start[candidate_schedule[i]] \
                          + PostDisasterLongTermSolution.compRepair

                schedule_time.append(end[candidate_schedule[i]])
                schedule_time.sort()

        # calculate the total recovery time
        trt += max(end.values())

        # compute the srt under current realization
        schedule_time = []
        pp = list(end.values())
        for i in range(len(pp)):
            if pp[i] not in schedule_time:
                schedule_time.append(pp[i])

        schedule_time = sorted(schedule_time)
        denominator = 0
        numerator = 0

        schedule = {}
        self_candidates = copy.deepcopy(self.candidates)

        # label whether a bridge has been repaired
        fg = {}
        for bridge in self_candidates:
            fg[bridge] = 0

        # to reduce the computation, just the performance at every 8-time
        # steps, you can adjust it
        inte = 1
        pl = len(schedule_time)

        for ii in range(pl % inte - 1 + inte, pl, inte):

            if ii > 0:

                # update the damage status of bridges
                for bridge in self_candidates:
                    if fg[bridge] == 0:
                        if end[bridge] <= schedule_time[ii]:
                            temp_bridge_damage_value[bridge] = 0
                            schedule[bridge] = [start[bridge], end[bridge]]
                            fg[bridge] = 1

                for i in range(len(self.arc_df)):
                    nod1 = self.node_df.loc[self.node_df['ID'] == self.arc_df['fromnode'][i], 'guid'].values[0]
                    nod2 = self.node_df.loc[self.node_df['ID'] == self.arc_df['tonode'][i], 'guid'].values[0]
                    self.network.edges[nod1, nod2]['Damage_Status'] = 0

                for key, val in temp_bridge_damage_value.items():
                    linknwid = self.bridge_df.loc[self.bridge_df['guid'] == key, 'linkID'].values[0]

                    nod_id1 = self.arc_df[self.arc_df['id'] == linknwid]['fromnode'].values[0]
                    nod1 = self.node_df.loc[self.node_df['ID'] == nod_id1, 'guid'].values[0]

                    nod_id2 = self.arc_df[self.arc_df['id'] == linknwid]['tonode'].values[0]
                    nod2 = self.node_df.loc[self.node_df['ID'] == nod_id2, 'guid'].values[0]

                    self.network.edges[nod1, nod2]['Damage_Status'] = val

                nx.get_edge_attributes(self.network, 'Damage_Status')

                # calculate the travel efficiency based on different
                # performance metrics based on travel time
                te = None
                if self.pm == 1:
                    te = TransportationRecoveryUtil.traveltime_freeflow(self.network)

                # based on WIPW
                elif self.pm == 0:
                    te = WIPW.tipw_index(self.network,
                                         self.all_ipw, self.path_adt)

                numerator += te * schedule_time[ii] * (schedule_time[ii]
                                                       - schedule_time[
                                                           ii - inte])
                aa = te
                denominator += te * (schedule_time[ii]
                                     - schedule_time[ii - inte])

        # calculate the skewness of the recovery trajectory
        try:
            srt += numerator / denominator
        except ZeroDivisionError as e:
            print(e)

        # the first objective assignment
        self.objectives[0] = trt

        # the second objective assignment
        self.objectives[1] = srt

        # repair schedule of bridge assignment
        self.sch[self.objectives[0], self.objectives[1]] = schedule

        if final == 0:
            return self.objectives[0], self.objectives[1]
        else:
            return self.objectives[0], self.objectives[1], \
                   self.sch[self.objectives[0], self.objectives[1]]

    def mutate(self):
        """
        Mutation operator
        """

        # label whether continuous the mutation operator, for each chromosome,
        # it just has one time mutation
        rec = True

        # implementation of mutation
        while rec:
            index1 = random.randint(0, len(self.candidates) - 1)
            index2 = random.randint(0, len(self.candidates) - 1)
            if index1 != index2:
                re = self.attributes[index1]
                self.attributes[index1] = self.attributes[index2]
                self.attributes[index2] = re
                rec = False



# Copyright (c) 2023 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


class InfrastructureArc(object):
    """
    This class models an arc in an infrastructure network

    Attributes
    ----------
    source : int
        Start (or head) node id
    dest : int
        End (or tail) node id
    layer : int
        The id of the layer of the network to which the arc belong
    failure_probability : float
        Failure probability of the arc
    functionality : bool
        Functionality state of the node
    repaired : bool
        If the arc is repaired or not
    flow_cost : float
        Unit cost of sending the main commodity through the arc
    reconstruction_cost : float
        Reconstruction cost of the arc
    capacity : float
        Maximum volume of the commodities that the arc can carry
    space : int
        The id of the geographical space where the arc is
    resource_usage : dict
        The dictionary that shows how many resource (of each resource type) is employed to repair the arc
    extra_com : dict
        The dictionary that shows flow_cost corresponding to commodities other than the main commodity
    is_interdep : bool
        If arc represent a normal arc (that carry commodity within a single layer) or physical interdependency between
        nodes from different layers
    """

    def __init__(self, source, dest, layer, is_interdep=False):
        self.source = source
        self.dest = dest
        self.layer = layer
        self.failure_probability = 0.0
        self.functionality = 1.0
        self.repaired = 1.0
        self.flow_cost = 0.0
        self.reconstruction_cost = 0.0
        self.resource_usage = {}
        self.capacity = 0.0
        self.space = 0
        self.extra_com = {}
        self.is_interdep = is_interdep

    def set_extra_commodity(self, extra_commodity):
        """
        This function initialize the dictionary for the extra commodities

        Parameters
        ----------
        extra_commodity : list
            List of extra commodities

        Returns
        -------
        None.

        """
        for ec in extra_commodity:
            self.extra_com[ec] = {'flow_cost': 0}

    def set_resource_usage(self, resource_names):
        """
        This function initialize the dictionary for resource usage per all resource types in the analysis

        Parameters
        ----------
        resource_names : list
            List of resource types

        Returns
        -------
        None.

        """
        for rc in resource_names:
            self.resource_usage[rc] = 0

    def in_space(self, space_id):
        """
        This function checks if the arc is in a given space or not

        Parameters
        ----------
        space_id :
            The id of the space that is checked

        Returns
        -------
            : bool
            Returns 1 if the arc is in the space, and 0 otherwise.
        """
        if self.space == space_id:
            return 1
        else:
            return 0



class BuildingData:
    def __init__(self, tract_id, lon, lat, structural, code_level, epsa_node_id, pwsa_node_id, tep_id, build_id_x,
                 epsa_id, pwsa_id, finance, ep_pw_id, occupation_code):
        self.tract_id = tract_id
        self.lon = lon
        self.lat = lat
        self.structural = structural
        self.code_level = code_level,
        self.epsa_node_id = epsa_node_id
        self.pwsa_node_id = pwsa_node_id
        self.tep_id = tep_id
        self.build_id_x = build_id_x
        self.epsa_id = epsa_id
        self.pwsa_id = pwsa_id
        self.finance = finance
        self.ep_pw_id = ep_pw_id
        self.occupation_code = occupation_code



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import math
import random
import sys

import networkx as nx
import numpy
from pyincore.utils.analysisutil import AnalysisUtil
from shapely.geometry import shape, LineString, MultiLineString

from pyincore import BaseAnalysis, HazardService, FragilityService, DataService, FragilityCurveSet
from pyincore import GeoUtil, NetworkUtil, NetworkDataset
from pyincore.models.dfr3curve import DFR3Curve


class TornadoEpnDamage(BaseAnalysis):
    """
    Computes electric power network (EPN) probability of damage based on a tornado hazard.
    The process for computing the structural damage is similar to other parts of the built environment.
    First, fragilities are obtained based on the hazard type and attributes of the network tower and network pole.
    Based on the fragility, the hazard intensity at the location of the infrastructure is computed. Using this
    information, the probability of exceeding each limit state is computed, along with the probability of damage.
    """

    def __init__(self, incore_client):
        self.hazardsvc = HazardService(incore_client)
        self.fragilitysvc = FragilityService(incore_client)
        self.datasetsvc = DataService(incore_client)
        self.fragility_tower_id = '5b201b41b1cf3e336de8fa67'
        self.fragility_pole_id = '5b201d91b1cf3e336de8fa68'

        # this is for deciding to use indpnode field. Not using this could be safer for general dataset
        self.use_indpnode = False
        self.nnode = 0
        self.highest_node_num = 0
        self.EF = 0
        self.nint = []
        self.indpnode = []
        self.mcost = 1435  # mean repair cost for single distribution pole
        self.vcost = (0.1 * self.mcost) ** 2
        self.sigmad = math.sqrt(
            math.log(self.vcost / (self.mcost ** 2) + 1))  # convert to gaussian Std Deviation to be used in logncdf
        self.mud = math.log((self.mcost ** 2) / math.sqrt(self.vcost + self.mcost ** 2))

        self.mcost = 400000  # mean repair cost for single transmission pole
        self.vcost = (0.1 * self.mcost) ** 2
        self.sigmat = math.sqrt(
            math.log(self.vcost / (self.mcost ** 2) + 1))  # convert to gaussian Std Deviation to be used in logncdf
        self.mut = math.log((self.mcost ** 2) / math.sqrt(self.vcost + self.mcost ** 2))

        self.tmut = 72  # mean repairtime for transmission tower in hrs
        self.tsigmat = 36  # std dev

        self.tmud = 5  # mean repairtime for poles in hrs
        self.tsigmad = 2.5

        self.totalcost2repairpath = []
        self.totalpoles2repair = []

        self.tornado_sim_field_name = 'SIMULATION'
        self.tornado_ef_field_name = 'EF_RATING'

        # tornado number of simulation and ef_rate
        self.nmcs = 0
        self.tornado_ef_rate = 0

        self.pole_distance = 38.1

        # node variables
        self.nodenwid_fld_name = "NODENWID"
        self.indpnode_fld_name = "INDPNODE"
        self.guid_fldname = 'GUID'

        # link variables
        self.tonode_fld_name = "TONODE"
        self.fromnode_fld_name = "FROMNODE"
        self.linetype_fld_name = "LINETYPE"

        # line type variable
        self.line_transmission = "transmission"
        self.line_distribution = "distribution"

        super(TornadoEpnDamage, self).__init__(incore_client)

    def run(self):
        network_dataset = NetworkDataset.from_dataset(self.get_input_dataset("epn_network"))
        tornado = self.get_input_hazard("hazard")
        tornado_id = self.get_parameter('tornado_id')
        if tornado is None and tornado_id is None:
            raise ValueError("Either tornado hazard object or tornado id must be provided")
        elif tornado_id is None:
            tornado_id = tornado.id

        tornado_metadata = self.hazardsvc.get_tornado_hazard_metadata(tornado_id)
        self.load_remote_input_dataset("tornado", tornado_metadata["datasetId"])
        tornado_dataset = self.get_input_dataset("tornado").get_inventory_reader()
        ds_results, damage_results = self.get_damage(network_dataset, tornado_dataset, tornado_id)

        self.set_result_csv_data("result", ds_results, name=self.get_parameter("result_name"))
        self.set_result_json_data(
            "metadata", damage_results, name=self.get_parameter("result_name") + "_additional_info")

        return True

    def get_damage(self, network_dataset, tornado_dataset, tornado_id):
        """

        Args:
            network_dataset (obj): Network dataset.
            tornado_dataset (obj): Tornado dataset.
            tornado_id (str): Tornado id.

        """
        link_dataset = network_dataset.links.get_inventory_reader()
        node_dataset = network_dataset.nodes.get_inventory_reader()
        self.set_tornado_variables(tornado_dataset)
        self.set_node_variables(node_dataset)

        # get fragility curves set - tower for transmission, pole for distribution
        fragility_set_tower = FragilityCurveSet(self.fragilitysvc.get_dfr3_set(self.fragility_tower_id))
        assert fragility_set_tower.id == self.fragility_tower_id
        fragility_set_pole = FragilityCurveSet(self.fragilitysvc.get_dfr3_set(self.fragility_pole_id))
        assert fragility_set_pole.id == self.fragility_pole_id

        # network test
        node_id_validation = NetworkUtil.validate_network_node_ids(
            network_dataset, self.fromnode_fld_name, self.tonode_fld_name, self.nodenwid_fld_name)
        if node_id_validation is False:
            sys.exit("ID in from or to node field doesn't exist in the node dataset")

        # getting network graph and node coordinates
        is_directed_graph = True
        link_filepath = link_dataset.path

        graph, node_coords = NetworkUtil.create_network_graph_from_link(
            link_filepath, self.fromnode_fld_name, self.tonode_fld_name, is_directed_graph)

        # reverse the graph to acculate the damage to next to node
        graph = nx.DiGraph.reverse(graph, copy=True)

        # check the connection as a list
        connection_sets = []
        if is_directed_graph:
            connection_sets = list(nx.weakly_connected_components(graph))
        else:
            connection_sets = list(nx.connected_components(graph))

        # check the first node of the each network line, this first node should lead each separated network
        # also convert connection set to list
        first_node_list = []
        connection_list = []
        for c in connection_sets:
            connection_list.append(list(c))
            first_node_list.append(list(c)[0])

        intersection_list = []
        poly_list = []
        totalcost2repair = []
        totalpoles2repair = []
        totaltime2repair = []

        # construct guid field
        guid_list = []
        nodenwid_list = []
        for node_feature in node_dataset:
            # get guid colum
            guid_fld_val = ''
            if self.guid_fldname.lower() in node_feature['properties']:
                guid_fld_val = node_feature['properties'][self.guid_fldname.lower()]
            elif self.guid_fldname in node_feature['properties']:
                guid_fld_val = node_feature['properties'][self.guid_fldname]
            guid_list.append(guid_fld_val)

            # get nodenwid colum
            nodenwid_fld_val = ''
            if self.nodenwid_fld_name.lower() in node_feature['properties']:
                nodenwid_fld_val = int(node_feature['properties'][self.nodenwid_fld_name.lower()])
            elif self.nodenwid_fld_name in node_feature['properties']:
                nodenwid_fld_val = int(node_feature['properties'][self.nodenwid_fld_name])
            nodenwid_list.append(nodenwid_fld_val)

        for z in range(self.nmcs):
            nodedam = [0] * self.nnode  # placeholder for recording number of damaged pole for each node
            noderepair = [0] * self.nnode  # placeholder for recording repair cost for each node
            poles2repair = [0] * self.nnode  # placeholder for recording total number of poles to repair
            cost2repairpath = [0] * self.nnode  # placeholder for recording total repair cost for the network
            time2repairpath = [0] * self.nnode  # placeholder for recording total repair time for the network
            nodetimerep = [0] * self.nnode
            hazardval = [[0]] * self.nnode  # placeholder for recording hazard values
            demandtypes = [[""]] * self.nnode  # placeholder for recording demand types
            demandunits = [[""]] * self.nnode  # placeholder for recording demand units

            # iterate link
            for line_feature in link_dataset:
                ndamage = 0  # number of damaged poles in each link
                repaircost = 0  # repair cost value
                repairtime = 0  # repair time value
                to_node_val = ""
                linetype_val = ""
                tor_hazard_values = [0]  # random wind speed in EF
                demand_types = [""]
                demand_units = [""]

                if self.tonode_fld_name.lower() in line_feature['properties']:
                    to_node_val = line_feature['properties'][self.tonode_fld_name.lower()]
                elif self.tonode_fld_name in line_feature['properties']:
                    to_node_val = line_feature['properties'][self.tonode_fld_name]

                if self.linetype_fld_name in line_feature['properties']:
                    linetype_val = line_feature['properties'][self.linetype_fld_name]
                elif self.linetype_fld_name.lower() in line_feature['properties']:
                    linetype_val = line_feature['properties'][self.linetype_fld_name.lower()]

                line = shape(line_feature['geometry'])

                # iterate tornado
                for tornado_feature in tornado_dataset:
                    resistivity_probability = 0  # resistivity value at the point of windSpeed
                    random_resistivity = 0  # random resistivity value between 0 and one

                    sim_fld_val = ""
                    ef_fld_val = ""

                    # get EF rating and simulation number column
                    if self.tornado_sim_field_name.lower() in tornado_feature['properties']:
                        sim_fld_val = int(tornado_feature['properties'][self.tornado_sim_field_name.lower()])
                    elif self.tornado_sim_field_name in tornado_feature['properties']:
                        sim_fld_val = int(tornado_feature['properties'][self.tornado_sim_field_name])

                    if self.tornado_ef_field_name.lower() in tornado_feature['properties']:
                        ef_fld_val = tornado_feature['properties'][self.tornado_ef_field_name.lower()]
                    elif self.tornado_ef_field_name in tornado_feature['properties']:
                        ef_fld_val = tornado_feature['properties'][self.tornado_ef_field_name]

                    if sim_fld_val == "" or ef_fld_val == "":
                        print("unable to convert tornado simulation field value to integer")
                        sys.exit(0)

                    # get Tornado EF polygon
                    # assumes that the polygon is not a multipolygon
                    poly = shape(tornado_feature['geometry'])
                    poly_list.append(poly)

                    # loop for ef ranges
                    for f in range(self.tornado_ef_rate):
                        npoles = 0  # number of poles in tornado ef box
                        poleresist = 0  # pole's resistance value
                        # setting EF rate value string to match in the tornado dataset's attribute table
                        ef_content = "EF" + str(f)

                        # compute the intersections between link line and ef polygon
                        # also figure out the length of the line that ovelapped with EF box
                        inter_length_meter = None
                        # compute the intersection between tornado polygon and line
                        if sim_fld_val == z and ef_fld_val.lower() == ef_content.lower():
                            if poly is not None and line is not None:
                                if poly.intersects(line):
                                    intersection = poly.intersection(line)
                                    any_point = None
                                    intersection_length = intersection.length
                                    if intersection.length > 0:
                                        # print(intersection.__class__.__name__)
                                        # calculate the length of intersected line
                                        # since this is a geographic, it has to be projected to meters to be calcuated
                                        inter_length_meter = GeoUtil.calc_geog_distance_from_linestring(intersection)
                                        if isinstance(intersection, MultiLineString):
                                            intersection_list.append(intersection)
                                            for inter_line in intersection.geoms:
                                                any_point = inter_line.centroid
                                                break
                                        elif isinstance(intersection, LineString):
                                            intersection_list.append(intersection)
                                            any_point = intersection.centroid

                                            # also, random point can be possible
                                            # by changing the following lines value 0.5
                                            # any_point = intersection.interpolate(0.5, normalized=True)

                                    if any_point is not None:
                                        # check if any_point is in the polygon
                                        if poly.contains(any_point) is False:
                                            # this is very hardly happen but should be needed just in case
                                            any_point = poly.centroid

                                    # check if the line is tower or transmission
                                    if linetype_val.lower() == self.line_transmission:
                                        fragility_set_used = fragility_set_tower
                                    else:
                                        fragility_set_used = fragility_set_pole

                                    values_payload = [{
                                        "demands": [x.lower() for x in fragility_set_used.demand_types],
                                        "units": [x.lower() for x in fragility_set_used.demand_units],
                                        "loc": str(any_point.coords[0][1]) + "," + str(any_point.coords[0][0])
                                    }]

                                    h_vals = self.hazardsvc.post_tornado_hazard_values(
                                        tornado_id, values_payload, self.get_parameter('seed'))
                                    tor_hazard_values = AnalysisUtil.update_precision_of_lists(
                                        h_vals[0]["hazardValues"])
                                    demand_types = h_vals[0]["demands"]
                                    demand_units = h_vals[0]["units"]
                                    hval_dict = dict()
                                    j = 0
                                    for d in h_vals[0]["demands"]:
                                        hval_dict[d] = tor_hazard_values[j]
                                        j += 1
                                    if isinstance(fragility_set_used.fragility_curves[0],
                                                  DFR3Curve):
                                        inventory_args = fragility_set_used.construct_expression_args_from_inventory(
                                            tornado_feature)
                                        resistivity_probability = \
                                            fragility_set_used.calculate_limit_state(
                                                hval_dict,
                                                inventory_type=fragility_set_used.inventory_type, **inventory_args)
                                    else:
                                        raise ValueError(
                                            "One of the fragilities is in deprecated format. This should not happen. "
                                            "If you are seeing this please report the issue.")

                                    # randomly generated capacity of each poles ; 1 m/s is 2.23694 mph
                                    poleresist = resistivity_probability.get('LS_0') * 2.23694
                                    npoles = int(round(inter_length_meter / self.pole_distance))
                                    repairtime_list = []

                                    mu = None
                                    sigma = None
                                    for k in range(npoles):
                                        repair_time = 0
                                        random_resistivity = random.uniform(0, 1)

                                        if random_resistivity <= poleresist:
                                            ndamage += 1
                                            # following codes can't be converted from matlab to python
                                            # however, the cross product <=3 or == 24 almost doesn't happen
                                            # since the time and cost differs when it is pole or tower,
                                            # this could be changed by see if it is tower or pole
                                            # if numpy.cross(k, z) <= 3 or numpy.cross(k, z) == 24:
                                            if linetype_val.lower() == self.line_transmission:
                                                mu = self.mut
                                                sigma = self.sigmat
                                                tmu = self.tmut
                                                tsigma = self.tsigmat
                                            else:
                                                mu = self.mud
                                                sigma = self.sigmad
                                                tmu = self.tmud
                                                tsigma = self.tsigmad

                                            repairtime_list.append(numpy.random.normal(tmu, tsigma))

                                    for k in range(ndamage):
                                        repaircost += numpy.random.lognormal(mu, sigma)

                                    # max of the repair time among different poles is taken
                                    # as the repair time for that line
                                    if len(repairtime_list) > 0:
                                        repairtime = max(repairtime_list)
                noderepair[to_node_val - 1] = repaircost
                nodedam[to_node_val - 1] = ndamage
                nodetimerep[to_node_val - 1] = repairtime
                hazardval[to_node_val - 1] = tor_hazard_values
                demandtypes[to_node_val - 1] = demand_types
                demandunits[to_node_val - 1] = demand_units

            # Calculate damage and repair cost based on network
            for i in range(len(first_node_list)):
                for j in range(len(connection_list[i])):
                    # print(connection_list[i][j], first_node_list[i])
                    pathij = list(nx.all_simple_paths(graph, connection_list[i][j], first_node_list[i]))
                    poler = 0
                    coster = 0
                    timer = []
                    # print(pathij)
                    if len(pathij) > 0:
                        for k in range(len(pathij)):
                            for var1 in range(len(pathij[k])):
                                poler = poler + nodedam[pathij[k][var1]]
                                coster = coster + noderepair[pathij[k][var1]]
                                # max of the time for different lines is taken as the repair time for that path.
                                # -- path is constituted of different lines.
                                timer.append(nodetimerep[pathij[k][var1]])
                    poles2repair[connection_list[i][j]] = poler
                    cost2repairpath[connection_list[i][j]] = coster
                    if len(timer) > 0:
                        time2repairpath[connection_list[i][j]] = max(timer)
                    else:
                        time2repairpath[connection_list[i][j]] = 0
            totalcost2repair.append(cost2repairpath)
            totalpoles2repair.append(poles2repair)
            totaltime2repair.append(time2repairpath)

        # create guid field from node dataset

        # calculate mean and standard deviation
        meanpoles = numpy.mean(numpy.asarray(totalpoles2repair), axis=0)
        stdpoles = numpy.std(numpy.asarray(totalpoles2repair), axis=0)
        meancost = numpy.mean(numpy.asarray(totalcost2repair), axis=0)
        stdcost = numpy.std(numpy.asarray(totalcost2repair), axis=0)
        meantime = numpy.mean(numpy.asarray(totaltime2repair), axis=0)
        stdtime = numpy.std(numpy.asarray(totaltime2repair), axis=0)

        # create result
        ds_results = []
        damage_results = []

        for i in range(len(meanpoles)):
            ds_result = dict()
            damage_result = dict()

            ds_result['guid'] = guid_list[i]
            ds_result["meanpoles"] = meanpoles[i]
            ds_result["stdpoles"] = stdpoles[i]
            ds_result["meancost"] = meancost[i]
            ds_result["stdcost"] = stdcost[i]
            ds_result["meantime"] = meantime[i]
            ds_result["stdtime"] = stdtime[i]
            ds_result['haz_expose'] = AnalysisUtil.get_exposure_from_hazard_values(hazardval[i], "tornado")

            damage_result['guid'] = guid_list[i]
            damage_result["fragility_tower_id"] = self.fragility_tower_id
            damage_result["fragility_pole_id"] = self.fragility_pole_id
            damage_result["hazardtype"] = "Tornado"
            damage_result['hazardvals'] = hazardval[i]
            damage_result['demandtypes'] = demandtypes[i]
            damage_result['demandunits'] = demandunits[i]

            ds_results.append(ds_result)
            damage_results.append(damage_result)

        return ds_results, damage_results

    """
    align coordinate values in a list as a single pair in order
    """

    def align_list_cooridnate(self, coord_list):
        coord_iterator = iter(coord_list)
        first = prev = next(coord_iterator)
        for coord in coord_iterator:
            yield prev, coord
            prev = coord

            # if it is polygon the following line is needed to close the polygon geometry
            # yield coord, first

    def set_tornado_variables(self, tornado_dataset):
        sim_num_list = []
        ef_rate_list = []

        for ef_poly in tornado_dataset:
            ef_string = ''
            if self.tornado_sim_field_name.lower() in ef_poly['properties']:
                sim_num_list.append(int(ef_poly['properties'][self.tornado_sim_field_name.lower()]))
            elif self.tornado_sim_field_name in ef_poly['properties']:
                sim_num_list.append(int(ef_poly['properties'][self.tornado_sim_field_name]))

            if self.tornado_ef_field_name.lower() in ef_poly['properties']:
                ef_string = ef_poly['properties'][self.tornado_ef_field_name.lower()]
            elif self.tornado_ef_field_name in ef_poly['properties']:
                ef_string = ef_poly['properties'][self.tornado_ef_field_name]
            # parse the number in EF and the format should be "EF0", "EF1", or something like it
            ef_rate_list.append(int(ef_string.lower().split("ef", 1)[1]))

        if len(sim_num_list) == 0 or len(ef_string) == 0:
            print("Could not convert tornado simulation value")
            sys.exit(0)

        self.nmcs = max(sim_num_list) + 1
        self.tornado_ef_rate = max(ef_rate_list) + 1

    def set_node_variables(self, node_dataset):
        i = 0

        for node_point in node_dataset:
            node_id = None
            indpnode_val = None
            if self.nodenwid_fld_name.lower() in node_point['properties']:
                node_id = int(node_point['properties'][self.nodenwid_fld_name.lower()])
            elif self.nodenwid_fld_name in node_point['properties']:
                node_id = int(node_point['properties'][self.nodenwid_fld_name])

            if self.use_indpnode is True:
                if self.indpnode_fld_name.lower() in node_point['properties']:
                    indpnode_val = int(node_point['properties'][self.indpnode_fld_name.lower()])
                elif self.indpnode_fld_name in node_point['properties']:
                    indpnode_val = int(node_point['properties'][self.indpnode_fld_name])

            if node_id is None and indpnode_val is None:
                print("problem getting the value")
                sys.exit(1)

            if self.use_indpnode is True:
                if indpnode_val > 0:
                    self.indpnode.append(node_id)
                else:
                    self.nint.append(node_id)
            else:
                self.nint.append(node_id)

            if node_id > self.highest_node_num:
                self.highest_node_num = node_id
            i += 1

        self.nnode = i

    def get_spec(self):
        return {
            'name': 'tornado-epn-damage',
            'description': 'tornado epn damage analysis',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'result dataset name',
                    'type': str
                },
                {
                    'id': 'tornado_id',
                    'required': False,
                    'description': 'Tornado hazard id',
                    'type': str
                },
                {
                    'id': 'seed',
                    'required': False,
                    'description': 'Initial seed for the tornado hazard value',
                    'type': int
                }
            ],
            'input_hazards': [
                {
                    'id': 'hazard',
                    'required': False,
                    'description': 'Hazard object',
                    'type': ["tornado"]
                },
            ],
            'input_datasets': [
                {
                    'id': 'epn_network',
                    'required': True,
                    'description': 'EPN Network Dataset',
                    'type': ['incore:epnNetwork'],
                },
                {
                    'id': 'tornado',
                    'required': False,
                    'description': 'Tornado Dataset',
                    'type': ['incore:tornadoWindfield'],
                }
            ],
            'output_datasets': [
                {
                    'id': 'result',
                    'parent_type': 'epn_network',
                    'description': 'CSV file of damages for electric power network by tornado',
                    'type': 'incore:tornadoEPNDamageVer3'
                },
                {
                    'id': 'metadata',
                    'parent_type': 'epn_network',
                    'description': 'Json file with information about applied hazard value and fragility',
                    'type': 'incore:tornadoEPNDamageSupplement'
                }
            ]
        }



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


# The order of import matters. You need to import module by order of dependency
from pyincore.client import Client
from pyincore.client import IncoreClient
from pyincore.client import InsecureIncoreClient
from pyincore.hazardservice import HazardService
from pyincore.utils.expressioneval import Parser
from pyincore.dataservice import DataService
from pyincore.utils.geoutil import GeoUtil
from pyincore.utils.networkutil import NetworkUtil
from pyincore.dataservice import DataService
from pyincore.fragilityservice import FragilityService
from pyincore.repairservice import RepairService
from pyincore.restorationservice import RestorationService
from pyincore.spaceservice import SpaceService
from pyincore.semanticservice import SemanticService
from pyincore.utils.analysisutil import AnalysisUtil
from pyincore.utils.popdisloutputprocess import PopDislOutputProcess
from pyincore.utils.cgeoutputprocess import CGEOutputProcess
from pyincore.utils.hhrsoutputprocess import HHRSOutputProcess
from pyincore.dataset import Dataset, InventoryDataset, DamageRatioDataset
from pyincore.models.fragilitycurveset import FragilityCurveSet
from pyincore.models.repaircurveset import RepairCurveSet
from pyincore.models.restorationcurveset import RestorationCurveSet
from pyincore.models.dfr3curve import DFR3Curve
from pyincore.models.mappingset import MappingSet
from pyincore.models.mapping import Mapping
from pyincore.models.networkdataset import NetworkDataset
from pyincore.models.hazard.hazarddataset import HazardDataset, HurricaneDataset, EarthquakeDataset, TsunamiDataset, \
    TornadoDataset, FloodDataset
from pyincore.models.hazard.hazard import Hazard
from pyincore.models.hazard.hurricane import Hurricane
from pyincore.models.hazard.flood import Flood
from pyincore.models.hazard.tsunami import Tsunami
from pyincore.models.hazard.earthquake import Earthquake
from pyincore.models.hazard.tornado import Tornado
from pyincore.models.units import Units
from pyincore.networkdata import NetworkData
from pyincore.baseanalysis import BaseAnalysis
import pyincore.globals

__version__ = pyincore.globals.PACKAGE_VERSION



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


import io
import json
import os
import re
import zipfile
import ntpath

import pyincore.globals as pyglobals
from pyincore import IncoreClient
from pyincore.utils import return_http_response
from urllib.parse import urljoin
import requests

logger = pyglobals.LOGGER


class DataService:
    """Data service client.

    Args:
        client (IncoreClient): Service authentication.

    """

    def __init__(self, client: IncoreClient):
        self.client = client
        self.base_url = urljoin(client.service_url, 'data/api/datasets/')
        self.files_url = urljoin(client.service_url, 'data/api/files/')
        self.base_earthquake_url = urljoin(client.service_url, 'hazard/api/earthquakes/')
        self.base_tornado_url = urljoin(client.service_url, 'hazard/api/tornadoes/')


    def get_dataset_metadata(self, dataset_id: str, timeout=(30, 600), **kwargs):
        """Retrieve metadata from data service. Dataset API endpoint is called.

        Args:
            dataset_id (str): ID of the Dataset.
            timeout (tuple[int,int]): Session timeout.
            **kwargs: A dictionary of external parameters.
        Returns:
            obj: HTTP response containing the metadata.

        """
        # construct url with service, dataset api, and id
        url = urljoin(self.base_url, dataset_id)
        r = self.client.get(url, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def get_dataset_files_metadata(self, dataset_id: str, timeout=(30, 600), **kwargs):
        """Retrieve metadata of all files associated with the dataset. Files API endpoint is called.

        Args:
            dataset_id (str): ID of the Dataset.
            timeout (tuple[int,int]): Session timeout.
            **kwargs: A dictionary of external parameters.

        Returns:
            obj: HTTP response containing the metadata.

        """
        url = urljoin(self.base_url, dataset_id + '/files')
        r = self.client.get(url, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def get_dataset_file_metadata(self, dataset_id: str, file_id: str, timeout=(30, 600), **kwargs):
        """Retrieve metadata of all files associated with the dataset. Files API endpoint is called.

        Args:
            dataset_id (str): ID of the Dataset.
            file_id (str): ID of the File.
            timeout (tuple[int,int]): Session timeout.
            **kwargs: A dictionary of external parameters.

        Returns:
            obj: HTTP response containing the metadata.

        """
        url = urljoin(self.base_url,
                                   dataset_id + "/files/" + file_id)
        r = self.client.get(url, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def get_dataset_blob(self, dataset_id: str, join=None, timeout=(30, 600), **kwargs):
        """Retrieve a blob of the dataset. Blob API endpoint is called.

        Args:
            dataset_id (str): ID of the Dataset.
            join (bool): Add join parameter if True. Default None.
            timeout (tuple[int,int]): Session timeout.
            **kwargs: A dictionary of external parameters.

        Returns:
            str: Folder or file name.

        """
        local_filename = None

        # add another layer of dataset id folder to differentiate datasets with the same filename
        cache_data_dir = os.path.join(self.client.hashed_svc_data_dir, dataset_id)

        # if cache_data_dir doesn't exist create one
        if not os.path.exists(cache_data_dir):
            os.makedirs(cache_data_dir)
            # for consistency check to ensure the repository hash is recorded in service.json
            self.client.create_service_json_entry()

            local_filename = self.download_dataset_blob(cache_data_dir, dataset_id, timeout=timeout, **kwargs )

        # if cache_data_dir exist, check if id folder and zip file exist inside
        else:
            for fname in os.listdir(cache_data_dir):
                if fname.endswith('.zip'):
                    local_filename = os.path.join(cache_data_dir, fname)
                    print('Dataset already exists locally. Reading from local cached zip.')
            if not local_filename:
                local_filename = self.download_dataset_blob(cache_data_dir, dataset_id, timeout=timeout, **kwargs)

        folder = self.unzip_dataset(local_filename)
        if folder is not None:
            return folder
        else:
            return local_filename

    def download_dataset_blob(self, cache_data_dir: str, dataset_id: str, join=None, timeout=(30, 600), **kwargs):
        # construct url for file download
        url = urljoin(self.base_url, dataset_id + '/blob')
        kwargs["stream"] = True
        if join is None:
            r = self.client.get(url, timeout=timeout, **kwargs)
        else:
            payload = {}
            if join is True:
                payload['join'] = 'true'
            elif join is False:
                payload['join'] = 'false'
            r = self.client.get(url, params=payload, timeout=timeout, **kwargs)
        # extract filename
        disposition = r.headers['content-disposition']
        fname = re.findall("filename=(.+)", disposition)

        local_filename = os.path.join(cache_data_dir, fname[0].strip('\"'))

        # download
        with open(local_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)

        return local_filename

    def get_datasets(self, datatype: str = None, title: str = None, creator: str = None, skip: int = None,
                     limit: int = None, space: str = None, timeout=(30, 600), **kwargs):
        """Function to get datasets. Blob API endpoint is called.

        Args:
            datatype (str): Data type of IN-CORE datasets. Examples: ergo:buildingInventoryVer5,
                ergo:census, default None
            title (str): Title of the dataset, passed to the parameter "title", default None
            creator (str): Dataset creator’s username, default None.
            skip (str): Skip the first n results, default None.
            limit (int): Limit number of results to return, default None.
            space (str): Name of space, default None.
            timeout (tuple[int,int]): Session timeout.
            **kwargs: A dictionary of external parameters.

        Returns:
            obj: HTTP response containing the metadata.

        """
        url = self.base_url
        payload = {}
        if datatype is not None:
            payload['type'] = datatype
        if title is not None:
            payload['title'] = title
        if creator is not None:
            payload['creator'] = creator
        if skip is not None:
            payload['skip'] = skip
        if limit is not None:
            payload['limit'] = limit
        if space is not None:
            payload['space'] = space

        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)
        # need to handle there is no datasets
        return return_http_response(r).json()

    def create_dataset(self, properties: dict, timeout=(30, 600), **kwargs):
        """Create datasets. Post API endpoint is called.

        Args:
            properties (dict): Metadata and files of the dataset to be created.
            timeout (tuple[int,int]): Session timeout.
            **kwargs: A dictionary of external parameters.
        Returns:
            obj: HTTP POST Response. Json of the dataset posted to the server.

        """
        payload = {'dataset': json.dumps(properties)}
        url = self.base_url
        kwargs["files"] = payload
        r = self.client.post(url, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def update_dataset(self, dataset_id, property_name: str,
                       property_value: str, timeout=(30, 600), **kwargs):
        """Update dataset. Put API endpoint is called.

        Args:
            dataset_id (str): ID of the Dataset.
            property_name (str): Property parameters such as name and value.
            property_value (str): Property parameters such as name and value.
            timeout (tuple[int,int]): Session timeout.
            **kwargs: A dictionary of external parameters.

        Returns:
            obj: HTTP PUT Response. Json of the dataset updated on the server.

        """
        url = urljoin(self.base_url, dataset_id)
        payload = {'update': json.dumps({"property name": property_name,
                                         "property value": property_value})}
        kwargs["files"] = payload
        r = self.client.put(url, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def add_files_to_dataset(self, dataset_id: str, filepaths: list, timeout=(30, 600), **kwargs):
        """Add files to the dataset. Post API endpoint is called.

        Args:
            dataset_id (str): ID of the Dataset.
            filepaths (list): Path to the files.
            timeout (tuple[int,int]): Session timeout.
            **kwargs: A dictionary of external parameters.

        Returns:
            obj: HTTP POST Response. Json of the files added to the dataset.

        """
        url = urljoin(self.base_url, dataset_id + "/files")
        listfiles = []
        for filepath in filepaths:
            file = open(filepath, 'rb')
            tuple = ('file', file)
            listfiles.append(tuple)
        kwargs['files'] = listfiles
        r = self.client.post(url, timeout=timeout, **kwargs)

        # close files
        for tuple in listfiles:
            tuple[1].close()

        return return_http_response(r).json()

    def add_files_to_network_dataset(self, dataset_id: str, filepaths: list,
                                     nodename: str, linkname: str, graphname: str, timeout=(30, 600), **kwargs):
        """Add files to the network dataset. Post API endpoint is called.

        Args:
            dataset_id (str): ID of the Dataset.
            filepaths (list): Path to the files.
            nodename (str): node shapefile name.
            linkname (str): link shapefile name.
            graphname (str): graph file name.
            timeout (tuple[int,int]): Session timeout.
            **kwargs: A dictionary of external parameters.

        Returns:
            obj: HTTP POST Response. Json of the files added to the dataset.

        """
        url = urljoin(self.base_url, dataset_id + "/files")
        listfiles = []
        linkname = os.path.splitext(linkname)[0]
        nodename = os.path.splitext(nodename)[0]
        graphname = os.path.splitext(graphname)[0]
        for filepath in filepaths:
            filename = os.path.splitext(ntpath.basename(filepath))[0]
            file = open(filepath, 'rb')
            bodyname = ''
            if filename == linkname:
                bodyname = 'link-file'
            if filename == nodename:
                bodyname = 'node-file'
            if filename == graphname:
                bodyname = 'graph-file'
            tuple = (bodyname, file)
            listfiles.append(tuple)
        kwargs["files"] = listfiles
        r = self.client.post(url, timeout=timeout, **kwargs)

        # close files
        for tuple in listfiles:
            tuple[1].close()

        return return_http_response(r).json()

    def delete_dataset(self, dataset_id: str, timeout=(30, 600), **kwargs):
        """Delete dataset. Delete API endpoint is called.

        Args:
            dataset_id (str): ID of the Dataset to be deleted.
            timeout (tuple[int,int]): Session timeout.
            **kwargs: A dictionary of external parameters.

        Returns:
            obj: HTTP DELETE Response. Json model of the delete action.

        """
        url = urljoin(self.base_url, dataset_id)
        r = self.client.delete(url, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def get_files(self, timeout=(30, 600), **kwargs):
        """Get all files. Files API endpoint is called.
        Args:
            timeout (tuple[int,int]): Session timeout.
            **kwargs: A dictionary of external parameters.
        Returns:
            obj: HTTP response containing the files.
            timeout (tuple[int,int]): Session timeout.
            **kwargs: A dictionary of external parameters.

        """
        url = self.files_url
        r = self.client.get(url, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def get_file_metadata(self, file_id: str, timeout=(30, 600), **kwargs):
        """Function to retrieve metadata of a file defined by id. Files API endpoint is called.

        Args:
            file_id (str): ID of the File.
            timeout (tuple[int,int]): Session timeout.
            **kwargs: A dictionary of external parameters.

        Returns:
            obj: HTTP response containing the metadata.

        """
        url = urljoin(self.files_url, file_id)
        r = self.client.get(url, timeout=timeout, **kwargs)
        return return_http_response(r).json()

    def get_file_blob(self, file_id: str, timeout=(30, 600), **kwargs):
        """Function to retrieve a blob of the file. Blob API endpoint is called.

        Args:
            file_id (str): ID of the Dataset.
            timeout (tuple[int,int]): Session timeout.
            **kwargs: A dictionary of external parameters.

        Returns:
            str: Local file name.

        """
        # construct url for file download
        url = urljoin(self.files_url, file_id + '/blob')
        kwargs["stream"] = True
        r = self.client.get(url, timeout=timeout, **kwargs)

        r = return_http_response(r)

        # extract filename
        disposition = r.headers['content-disposition']
        fname = re.findall("filename=(.+)", disposition)

        # construct local directory and filename
        if not os.path.exists('data'):
            os.makedirs('data')
        local_filename = os.path.join('data', fname[0].strip('\"'))

        # download
        with open(local_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)

        return local_filename

    def unzip_dataset(self, local_filename: str):
        """Unzip the dataset zip file.

        Args:
            local_filename (str): Name of the Dataset.

        Returns:
            str: Folder name with unzipped files.

        """
        foldername, file_extension = os.path.splitext(local_filename)
        # if it is not a zip file, no unzip
        if not file_extension.lower() == '.zip':
            print('It is not a zip file; no unzip')
            return None
        # check the folder existance, no unzip
        if os.path.isdir(foldername):
            print('Unzipped folder found in the local cache. Reading from it...')
            return foldername
        os.makedirs(foldername)

        zip_ref = zipfile.ZipFile(local_filename, 'r')
        zip_ref.extractall(foldername)
        zip_ref.close()
        return foldername

    def get_shpfile_from_service(self, fileid, dirname, timeout=(30, 600), **kwargs):
        """Function to obtain a shape file from Data service.

        Args:
            fileid (str):  ID of the Shape file.
            dirname (str): Directory the files are being extracted.
            timeout (tuple[int,int]): Session timeout.
            **kwargs: A dictionary of external parameters.

        Returns:
            str: Filename with shape files.

        """
        request_str = self.base_url + fileid
        request_str_zip = request_str + '/blob'

        # obtain file name
        r = self.client.get(request_str, timeout=timeout, **kwargs)
        r = return_http_response(r).json()
        first_filename = r["fileDescriptors"][0]["filename"]
        filename = os.path.splitext(first_filename)[0]
        kwargs["stream"] = True

        r = self.client.get(request_str_zip, timeout=timeout, **kwargs)
        r = return_http_response(r)
        z = zipfile.ZipFile(io.BytesIO(r.content))
        z.extractall(dirname)
        # print(r.status_code)

        return filename

    def get_tornado_dataset_id_from_service(self, fileid, timeout=(30, 600), **kwargs):
        """Function to obtain a tornado dataset Id from Data service.

        Args:
            fileid (str): ID of the Tornado file.
            timeout (tuple[int,int]): Session timeout.
            **kwargs: A dictionary of external parameters.

        Returns:
            obj: HTTP response containing the tornado Dataset ID.

        """
        # dataset
        request_str = self.base_tornado_url + fileid
        r = self.client.get(request_str, timeout=timeout, **kwargs)

        return return_http_response(r).json()["tornadoDatasetId"]

    def search_datasets(self, text: str, skip: int = None, limit: int = None, timeout=(30, 600), **kwargs):
        """Function to search datasets.

        Args:
            text (str): Text to search by, passed to the parameter "text".
            skip (int):  Skip the first n results, passed to the parameter "skip". Dafault None.
            limit (int):  Limit number of results to return. Passed to the parameter "limit". Dafault None.
            timeout (tuple[int,int]): Session timeout.
            **kwargs: A dictionary of external parameters.

        Returns:
            obj: HTTP response with search results.

        """
        url = urljoin(self.base_url, "search")
        payload = {"text": text}
        if skip is not None:
            payload['skip'] = skip
        if limit is not None:
            payload['limit'] = limit

        r = self.client.get(url, params=payload, timeout=timeout, **kwargs)

        return return_http_response(r).json()
